/******************************************************************
 * CLASS:			MiscArrowButtonCell
 *
 *	See the header for more information on this class.
 *
 * This object is included in the MiscKit by permission from the author
 * and its use is governed by the MiscKit license, found in the file
 * "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
 * for a list of all applicable permissions and restrictions.
 *******************************************************************/

#import "wraps.h"
#import <AppKit/psops.h>
#import <InterfaceBuilder/InterfaceBuilder.h>
#import "MiscArrowButtonCell.h"

// These #defines are used for versioning instances of this class, so that
// you will always be able to read any previously archived version. If you
// make changes to the class's ivars, bump the version number up one and 
// make the appropriate changes to the read: method.

#define MISC_ABC_VERSION 1
#define MISC_ABC_CLASSNAME "MiscArrowButtonCell"


@implementation MiscArrowButtonCell

+ (void)initialize
{
	if (self == [MiscArrowButtonCell class])
		[self setVersion: MISC_ABC_VERSION];
	
	return;
}



// The initializing methods just call the designated initializer
// and give some default values.

- init
{
	return [self initTextCell:@"Left" altTitle:@"Right"];
}



- initTextCell:(NSString *)aString
{
	return [self initTextCell:aString altTitle:@"Right"];
}



// The desinated initializer for this class.

- initTextCell: (NSString *)aString altTitle: (NSString *)altString
{
	[super initTextCell:aString];
	
	[self setAlternateTitle:altString];
	[self setState:0];
	[self setBordered:NO];
	[self setShowsStateBy:NSNoCellMask];
	[self setHighlightsBy:NSNoCellMask];
	[self setArrowAlignment: MISC_ARROW_ABSOLUTE];
	[self setType:NSToggleButton];
	
	return self;
}



// Since the stringValue in Button does not do much, it now
// returns the currently selected text.

- (NSString *)stringValue
{
	if ([self state] == 0)
		return [self title];
	else
		return [self alternateTitle];
}



// I believe this method is supposed to calculate the minimum size needed to 
// fit the currently displayed contents  (contents, altContents, and the arrow)
// in the cellframe.

- (NSSize)cellSizeForBounds:(NSRect)theRect
{
  
    NSSize theSize;
    float  contentsWidth = [[self font] widthOfString:[self title]];
  float  altContentsWidth = [[self font] widthOfString:[self alternateTitle]];
  
	// first calculate the width needed to draw all text and the arrow
	  
  	theSize.width = 0.0;
	if ([self arrowAlignment] == MISC_ARROW_RELATIVE)
	{
		theSize.width += contentsWidth;
		theSize.width += altContentsWidth;
	 }
	else	// absolute alignment
	{
		if (contentsWidth > altContentsWidth)
			theSize.width += contentsWidth * 2.0;
		else
			theSize.width += altContentsWidth * 2.0;
	 }
	 
	theSize.width += cellHeight + 10.0;					
	
	// now the height (which will usually be cellHeight)
	
	theSize.height = cellHeight;
	
	if ([ [self font] pointSize] > cellHeight)
		theSize.height = [ [self font] pointSize];
	 		
	return theSize;
}



// As far as I can tell (which may not be all that far) is that this 
// function is only used by the Button's IBEditor to tell how large
// the editor should be.

- (NSRect)titleRectForBounds:(NSRect)theRect
{
  float  size = [ [self font] pointSize];
  float  theY;
  float  maxWidth = theRect.size.width/2.0;
				
	// this is a hack so you can double click on the altTitle in IB

	if ([self state])
		[self getAltTitleRect:&theRect];

	[self setAlignment:NSLeftTextAlignment];	// used for the IBEditor
		
	theY = theRect.origin.y + theRect.size.height/2.0 - size/2.0;
	
	if ([self arrowAlignment] == MISC_ARROW_RELATIVE)
	  	maxWidth = [[self font] widthOfString:[self title]] + 10.0;  	
		
	theRect = NSMakeRect(theRect.origin.x, theY-2.0, maxWidth, size+2.0);		   	

	return theRect;
}



// If the state is 1 (altContents selected), then this method will be
// called. It returns the location for editor to appear when editing 
// the altContents.

- getAltTitleRect: (NSRect *)theRect
{
  float  size = [ [self font] pointSize];
  float  newX, newY, newWidth;
	
	[self setAlignment:NSRightTextAlignment];	// used for the IBEditor

	newX = theRect->origin.x + theRect->size.width/2.0; 			 
	newY = theRect->origin.y + theRect->size.height/2.0 - size/2.0;
	newWidth = theRect->size.width/2.0;
	
	*theRect = NSMakeRect(newX, newY-2.0, newWidth, size+2.0);		   	
	
	return self;
}



// New method to set the alignment of the arrow. MISC_ARROW_ABSOLUTE aligns
// the arrow in the center of the cellFrame, where MISC_ARROW_RELATIVE
// centers the arrow between the contents and altContents.

- (void)setArrowAlignment: (int)alignment
{
    arrowAlignment = alignment;
}



- (int)arrowAlignment
{
	return arrowAlignment;
}



// This method is used to draw only the parts of the cell that
// do change when the state changes. Therefore, only the black arrow
// and the text are drawn here.

- (void)drawInteriorWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{
  float  gray;
  float  size = [ [self font] pointSize];
		    
	// draw the arrow

	if ([self arrowAlignment] == MISC_ARROW_RELATIVE)
	{
	  NSRect  relativeFrame;
	  float  contentsWidth = [[self font] widthOfString:[self title]];
  	  float  altContentsWidth = [[self font] widthOfString:[self alternateTitle]];

		// since it the arrow is drawn relative to the text, a little
		// calculation is needed
		
		relativeFrame.origin.x = cellFrame.origin.x + contentsWidth;
		relativeFrame.size.width = cellFrame.size.width - 
				(contentsWidth + altContentsWidth);

		PSABdrawarrow (relativeFrame.origin.x, cellFrame.origin.y + 3.0, 
				relativeFrame.size.width, cellFrame.size.height - 6.0, 
				(int)[self state], [self isEnabled]);
	 }
	else
		// draw the arrow in the center of the cellFrame
		
		PSABdrawarrow (cellFrame.origin.x, cellFrame.origin.y + 3.0, 
				cellFrame.size.width, cellFrame.size.height - 6.0, 
				(int)[self state], [self isEnabled]);
	
	// draw the left and right hand text
	
	[[self font] set];
			
	if ([[self title] cString] != NULL)
	{	
	  float  theY;
	  			
		if ([self state] || ![self isEnabled])
			gray = NSDarkGray;
		else 
			gray = 0.0;

		// calculate the placement of the contents and print it
		
		theY = cellFrame.origin.y + cellFrame.size.height/2.0 + size/3.0;			

		PSABshowstring (cellFrame.origin.x + 3.0, theY, gray, [[self title] cString]);
	 }
		 
	if ([[self alternateTitle] cString] != NULL)
	{	
	  float  theX, theY;
	  float  strWidth = [[self font] widthOfString:[self alternateTitle]];
	  			
		if ([self state] && [self isEnabled])
			gray = 0.0;
		else
			gray = NSDarkGray;
		
		// calculate the placement of the altContents and print it
		
		theX = cellFrame.origin.x+(cellFrame.size.width - strWidth - 3.0);
		theY = cellFrame.origin.y + cellFrame.size.height/2.0 + size/3.0;
					
		PSABshowstring (theX, theY, gray, [[self alternateTitle] cString]);		 	
	 }
}



// This part of the drawing displays only the parts that don't change
// often, which includes the diamond that encloses the arrow, and the
// border when I get around to adding one. 

- (void)drawWithFrame:(NSRect)cellFrame inView:(NSView *)controlView
{		  
	// if transparent draw the background white (when in IB) and 
	// same as the background when in an app (or testing interface)
	
	if ([self isTransparent])
	{	
		if ([NSApp respondsToSelector:@selector(isTestingInterface)])
			if ([NSApp isTestingInterface])
				PSsetgray (NSLightGray);
			else
				PSsetgray (1.0);
		else
			PSsetgray (NSLightGray);
		
		NSRectFill(cellFrame);

		return;
	 }
	
	// set cellHeight since some of the calculations that are in other
	// methods have to know the height of the cell (this is probably the
	// wrong way to go about this)
	
	cellHeight = cellFrame.size.height;
	
	// draw the border or bezel, then call drawInside
	
	if ([self isBordered])
		NSDrawButton(cellFrame , cellFrame);
		
	PSgsave();

	// have to flip drawing if we are drawing into a flipped view
			
	if ([controlView isFlipped])
		PSABflipme (cellFrame.size.height + (cellFrame.origin.y * 2.0));
	
	if ([self arrowAlignment] == MISC_ARROW_RELATIVE)
	{
	  NSRect  relativeFrame;
	  float  contentsWidth = [[self font] widthOfString:[self title]];
  	  float  altContentsWidth = [[self font] widthOfString:[self alternateTitle]];

		relativeFrame.origin.x = cellFrame.origin.x + contentsWidth;
		relativeFrame.size.width = cellFrame.size.width - 
				(contentsWidth + altContentsWidth);
		
		PSABdrawdiamond (relativeFrame.origin.x, cellFrame.origin.y + 3.0,
				relativeFrame.size.width, cellFrame.size.height - 6.0);
	 }
	else
		PSABdrawdiamond (cellFrame.origin.x, cellFrame.origin.y + 3.0, 
			cellFrame.size.width, cellFrame.size.height - 6.0);
		
	PSgrestore();				

	// draw the rest of the cell
				
	[self drawInteriorWithFrame:cellFrame inView:controlView];
}



// Archiving methods

- (id)initWithCoder:(NSCoder *)aDecoder
{
	int  version;
	
	[super initWithCoder:aDecoder];
	version = [aDecoder versionForClassName:[NSString stringWithCString:MISC_ABC_CLASSNAME]];
	
	switch (version)
	{
	  case 0:
	  	// First version had no ivars, but set any new ivars to a useable
		// value.
	  	arrowAlignment = MISC_ARROW_ABSOLUTE;	
	  	break;
	  case 1:
	  	// Current version.
		[aDecoder decodeValueOfObjCType:@encode(unsigned int) at:&arrowAlignment];
		break;
	  default:
	  	break;
	 }

	return self;
}



- (void)encodeWithCoder:(NSCoder *)aCoder
{
	[super encodeWithCoder:aCoder];
	
	// You can also keep versioning here just in case you would like 
	// to back to an older version, but I don't.
	[aCoder encodeValueOfObjCType:@encode(unsigned int) at:&arrowAlignment];
}

@end


/******************************************************************
  CHANGES:
  	1. Now use setArrowAlignment/arrowAlignment to set/get whether the
 	 the arrow is relative or absolutely aligned. Added arrowAlignment 
	 instance var.
   October 1, 1994:
	2. Added archive versioning (+initialize).
	
 *******************************************************************/
 

 