/****************************************************************************
  CLASS:				MiscGaugeCell
  
  For more information, see the interface file.
  
  Since most of the code from this class came from the GaugeView class, here
  is the original comments for your reading pleasure:
   
  GaugeView.m, analog gauge view
  Author: Bruce Blumberg, NeXT Developer Support Group.
  Originally written for 0.6 mid 1988, modified for 1.0 by Ali Ozer.
  Redesigned for 2.0 by Julie Zelenski, NeXT Developer Support
 
  Subclass of view to implement a simple round analog gauge. You can set the 
  minimum, maximum value, start angle, angle range, title, font, and more.  
  It is a pretty generic round gauge view, if you ever have need for one.
 
  You may freely copy, distribute and reuse the code in this example.  
  NeXT disclaims any warranty of any kind, expressed or implied, as to 
  its fitness for any particular use.

  This object is included in the MiscKit by permission from the author
  and its use is governed by the MiscKit license, found in the file
  "LICENSE.rtf" in the MiscKit distribution.  Please refer to that file
  for a list of all applicable permissions and restrictions.
 ****************************************************************************/

//#import <AppKit/AppKit.h>
#import "Gauge.h"				// For the pswraps.
#import "MiscGaugeCell.h"

// Used for class versioning. If you happen to change the ivars, then 
// bump the version up and update the read:/write: methods.
#define MISCGAUGECELL_VERSION 0
#define MISCGAUGECELL_NAME @"MiscGaugeCell"


@implementation MiscGaugeCell 

+ (void)initialize
{
	if (self == [MiscGaugeCell class])
		// Set the class's version for archiving purposes.
		[self setVersion: MISCGAUGECELL_VERSION];
}


- init
{
	if ([super init] == nil)
		return nil;
	
	cache = [ [NSImage alloc] init];   
	radius = 0.0;
	center.x = center.y = 0.0;
	startAngle = angleRange = degreesPerUnit = 0.0;
	tickInterval = 0;
	gaugeColor = [NSColor darkGrayColor];
	textColor = [NSColor blackColor];
//	backgroundColor = NXConvertGrayToColor (NX_LTGRAY);
	minValue = 0.0; 
	maxValue = 1.0;
	value = 0.0;
	strValue = NULL;
	tickRatio = 0.65;
	handRatio = 0.6;
    degreesPerUnit = angleRange/(maxValue-minValue);
	[self setType: NSTextCellType];	// Need this so I can use [self font].
	[self setTitle:@"Gauge"];
	[self setTitlePosition: NSAtTop];
	titleFont = [self font]; 

	return self;
}


- (void)dealloc
{
	if (strValue != NULL)
		free (strValue);
       	[cache release];
	{ [super dealloc]; return; };
}


- (float)maxValue { return maxValue; }
- (float)minValue { return minValue; }
- (float)floatValue	{ return value; }
- (double)doubleValue { return (double)value; }
- (int)intValue { return (int)value; }


- (NSString *)stringValue 
{ 
	if (strValue != NULL)
		free (strValue);
	 
	strValue = malloc((20) * sizeof(char));
	sprintf (strValue, "%f", value);
	return [NSString stringWithCString:strValue];
}


- (void)setMaxValue: (float)max
{
	if ([self maxValue] == max)
		return;
		
	maxValue = max;
	if ([self minValue] >= [self maxValue])
		[self setMinValue: max-1];
	if ([self floatValue] > maxValue)
		[self setFloatValue: maxValue];

    degreesPerUnit = angleRange/(maxValue-minValue);
}


- (void)setMinValue: (float)min
{
	if ([self minValue] == min)
		return;
		
	minValue = min;
	if ([self minValue] >= [self maxValue])
		[self setMaxValue: min+1];
	if ([self floatValue] < minValue)
		[self setFloatValue: minValue]; 

    degreesPerUnit = angleRange/(maxValue-minValue);
}


- (void)setFloatValue: (float)val
{
	if (val < [self minValue])
		val = minValue;
	if (val > [self maxValue])
		val = maxValue;
	value = val;
}


- (void)setDoubleValue: (double)val
{
	[self setFloatValue: (float)val];
}


- (void)setIntValue: (int)val
{
	[self setFloatValue: (float)val];
}


- (void)setStringValue:(NSString *)val
{
	if ([val cString] == NULL)
		[self setFloatValue: 0.0];
	else
		[self setDoubleValue: atof([val cString])];
}


//- (NXColor)backgroundColor { return backgroundColor; }
- (NSColor *)gaugeColor { return gaugeColor; }
- (NSColor *)textColor { return textColor; }

/*
- (void)setBackgroundColor: (NXColor)color
{
	if (NXEqualColor (color, backgroundColor) == NO)
		backgroundColor = color;
}
*/

- (void)setGaugeColor:(NSColor *)color
{
	if ([color isEqual:gaugeColor] == NO)
		gaugeColor = color; 
}


- (void)setTextColor: (NSColor *)color
{
	if ([color isEqual:textColor] == NO)
		textColor = color;
}

/*
- (float)backgroundGray	
{
	float gray;
	NXConvertColorToGray (backgroundColor, &gray);
 	return gray; 
}
*/

- (float)gaugeGray 
{ 
	float gray;
	[[gaugeColor colorUsingColorSpaceName:NSCalibratedWhiteColorSpace] getWhite:&gray alpha:NULL];
 	return gray; 
}


- (float)textGray 
{ 
	float gray;
	[[textColor colorUsingColorSpaceName:NSCalibratedWhiteColorSpace] getWhite:&gray alpha:NULL];
 	return gray; 
}

/*
- (void)setBackgroundGray: (float)gray
{
	[self setBackgroundColor: NXConvertGrayToColor(gray)];
}
*/

- (void)setGaugeGray:(float)gray
{
	[self setGaugeColor:[NSColor colorWithCalibratedWhite:gray alpha:1.0]]; 
}


- (void)setTextGray: (float)gray
{
	[self setTextColor: [NSColor colorWithCalibratedWhite:gray alpha:1.0]];
}


- (float)startAngle { return startAngle; }
- (float)angleRange	{ return angleRange; }
- (int)tickInterval { return tickInterval; }
- (float)tickRatio { return tickRatio; }
- (float)handRatio { return handRatio; }


- (void)setStartAngle:(float)newValue
{
	// Sets start angle for gauge, which is the angle of the arm when at the 
	// minimum value. needRedraw is set to indicate that face image must be 
	// redrawn. 0 degrees is straight right, with increasing numbers going
	// counter clockwise.
	if (newValue < 0.0)
		newValue = 0.0;
	if (newValue > 360.0)
		newValue = 360.0;
    startAngle = newValue; 
}


- (void)setAngleRange:(float)newValue
{
	// Sets angle range for gauge, which is the sweep of the arm from minimum
	// value to maximum value.  The value cannot exceed 360 degrees (a full
	// revolution).  Recalculates degreesPerUnit which is used to 
	// determine interval of ticks and labels. needRedraw is set to indicate 
	// that face image must be redrawn.
    if (newValue > 360.0) 
		newValue = 360.0;
	if (newValue < 0.0) 
		newValue = 0.0;
	
    angleRange = newValue;
    degreesPerUnit = angleRange/(maxValue-minValue); 
}


- (void)setTickInterval:(int)newValue
{
	// Sets tick interval for gauge, which is number of units between tick
	// marks on gauge face.  needRedraw is set to indicate that 
	// face image must be redrawn.
	if (newValue < 1)
		newValue = 1;
    tickInterval = newValue; 
}


- (void)setTickRatio:(float)newRatio
{
	// Make sure new tick ratio is between 0.0 (no ticks) and 1.0.
	if (newRatio < 0.0)
		newRatio = 0.0;
	if (newRatio > 1.0)
		newRatio = 1.0;
	tickRatio = newRatio; 
}


- (void)setHandRatio:(float)newRatio
{
	// Make sure new hand ratio is between 0.2 and 0.9.
	if (newRatio < 0.2)
		newRatio = 0.2;
	if (newRatio > 0.9)
		newRatio = 0.9;
	handRatio = newRatio; 
}


- titleFont { return titleFont; }
- (NSString *)title { return [super stringValue]; }
- (NSTitlePosition)titlePosition { return titlePosition; }

- (void)setTitleFont: newFont
{
	if ([self titleFont] != newFont)
		titleFont = newFont;
}


- (void)setTitle:(NSString *)newTitle
{
	// We are making use of the ivar contents inherited from Cell.
	// Since we overrode stringValue and setStringValue:, we have
	// to use Cell's implementation.
	[super setStringValue:newTitle];
}	
	

- (void)setTitlePosition: (NSTitlePosition)newPos
{
	if (newPos == NSAtTop || newPos == NSAtBottom)
		titlePosition = newPos;
}


- (NSSize)cellSizeForBounds:(NSRect)rect
{
	// Return the minimum size that the cell will fit in.
	
    NSSize size;
    size.width = NSWidth(rect);
	size.height = NSHeight(rect);

	if (size.height < size.width) 
		size.width = size.height;
	else 
		size.height = size.width;

	return size;
}


- (void)drawWithFrame:(NSRect)rect inView:controlView
{
    center.x = NSWidth(rect)/2.0;
    center.y = NSHeight(rect)/2.0;
	if (NSWidth(rect) > NSHeight(rect))
    	radius = (NSHeight(rect)/2.0) - 8.0;
	else
		radius = (NSWidth(rect)/2.0) - 8.0;

	// Redraws the face in the cached image.		
	[self drawFace:rect];
	// Draws the gauge itself. 
	[self drawInteriorWithFrame:rect inView:controlView];
}


- (void)drawInteriorWithFrame:(NSRect)rect inView:controlView
{
	// Composites the "face" of the gauge, then draws the hand.
	// This method will also work inside a flipped view (like a Matrix).
	NSPoint corner = {0.0, 0.0};

	// If view is flipped composite from lower left corner.
	if ([controlView isFlipped])
		corner.y = NSHeight(rect);

	// Composite the gauge face into the view.
	[cache compositeToPoint:corner operation:NSCompositeSourceOver];
	// Draw the hand.
	[self drawHand: &rect flipped: [controlView isFlipped] ];	
}


- (void)drawHand: (const NSRect *)rect flipped: (BOOL)viewFlipped
{
	// If you want the gauge to have a different look you should only
	// have to override drawFace: and drawHand:. The passed rect
	// is the same rect that is passed to drawInside:.
    float valueAngle;	

    valueAngle = startAngle - degreesPerUnit*(value-minValue);
	
	if (viewFlipped == YES)
	{
		PSgsave();
		PSWflipme (NSHeight(*rect));
	 }
	
	[[self textColor] set];
    PSWdrawHand(center.x,center.y,valueAngle, radius*handRatio);

	if (viewFlipped == YES)
		PSgrestore();
}

	
- (void)drawFace:(NSRect)rect 
{	
	// This method draws the gauge face image in an NXImage.  It erases
	// the background, draws the circular border, draws the tick marks and 
	// labels them appropriately. Since we are drawing into an NXImage
	// we don't care whether the view we are compositing into is flipped or 
	// not.
    float angle, angleIncrement;
    int number;
    NSSize string;
    NSPoint pt;
    char numString[10];
    NSSize cacheSize;
	NSFont *cellsFont = [self font];
	NSFont *flippedFont;
	float titleY;
	
	cacheSize.width = NSWidth(rect);
	cacheSize.height = NSHeight(rect);
	[cache setSize:cacheSize];
	
	[cache lockFocus];

	// Do the entire rect in our background color.
//	NXSetColor ([self backgroundColor]);
	PSsetalpha (0.0);
    NSRectFill(rect);
	PSsetalpha (1.0);

	// Draw the gauge face.
	[[self gaugeColor] set];
    PSWdrawBorder(center.x, center.y, radius);
    angleIncrement = angleRange/((maxValue-minValue)/(float)tickInterval); 
	if (tickRatio > 0.0)
	{
		[[self textColor] set];
    	PSWdrawTicks(center.x, center.y, radius*tickRatio, angleIncrement/2.0, 
						startAngle, startAngle+angleRange); 
	 }

	if ([[self title] cString] != NULL)
	{
		// Makes sure that the title font if using the NX_IDENTITYMATRIX.	 
		flippedFont = [NSFont fontWithName:[titleFont fontName] 
			size:[titleFont pointSize]];
		[flippedFont set];     
	
		// Spit out the title.
		string.height = [flippedFont pointSize];
		string.width = [flippedFont widthOfString:[self title]];
		if ([self titlePosition] == NSAtTop)
		    titleY = center.y+(NSHeight(rect)/10.0);
		else
		    titleY = center.y-(NSHeight(rect)/12.0)-string.height;
		
		[[self textColor] set];	
		PSWdrawString((NSWidth(rect)-string.width)/2, titleY, 
						[[self title] cString]);
	 }
	 
	// Make sure the cell's font (for the numbers) is also using
	// the NX_IDENTITYMATRIX.	 
	flippedFont = [NSFont fontWithName:[cellsFont fontName] 
		size:[cellsFont pointSize]];
	[flippedFont set];     
    string.height = [flippedFont pointSize];
    
	// Spit out all the numbers.
    number =  minValue;
	[[self textColor] set];
    for(angle=startAngle;angle>=startAngle-angleRange;angle -= angleIncrement){
        sprintf(numString,"%d",number);
		string.width = [cellsFont widthOfString:[NSString stringWithCString:numString]];
		pt.x = cos(angle/57.3)*(radius-1.0-string.width/2)+ center.x; 
		pt.y = sin(angle/57.3)*(radius-1.0-string.height/2) + center.y ;
		PSWdrawString(pt.x-(string.width/2), pt.y-(string.height/2.5),
			numString);
		number += tickInterval;
    }
    
    [cache unlockFocus]; 
}	


- (void)highlight:(BOOL)flag withFrame:(NSRect)cellFrame inView:aView
{
	// We don't want highlighting to occur.
}


- (id)initWithCoder:(NSCoder *)aDecoder
{
	int version;
	
	[super initWithCoder:aDecoder];
	version = [aDecoder versionForClassName:MISCGAUGECELL_NAME];
	switch (version)
	{
		case MISCGAUGECELL_VERSION:
			[aDecoder decodeValuesOfObjCTypes:"ffiff", &startAngle, &angleRange, 
							&tickInterval, &tickRatio, &handRatio];
			[aDecoder decodeValuesOfObjCTypes:"fffs", &minValue, &maxValue, &value,
							&titlePosition];
//			backgroundColor = NXReadColor (stream);

			gaugeColor = [[aDecoder decodeObject] retain];
			textColor = [[aDecoder decodeObject] retain];
			titleFont = [[aDecoder decodeObject] retain];
			break;
			
		default:
			NSLog(@"Could not read typed stream for class %s, version %d", [MISCGAUGECELL_NAME cString], version);
			break;
	 }

        cache = [ [NSImage alloc] init];
        strValue = NULL;
	degreesPerUnit = angleRange/(maxValue-minValue);

	return self;
}


- (void)encodeWithCoder:(NSCoder *)aCoder
{
	[super encodeWithCoder:aCoder];
	[aCoder encodeValuesOfObjCTypes:"ffiff", &startAngle, &angleRange, 
					&tickInterval, &tickRatio, &handRatio];
	[aCoder encodeValuesOfObjCTypes:"fffs", &minValue, &maxValue, &value,
					&titlePosition]; 
//	NXWriteColor (stream, backgroundColor);
	[aCoder encodeObject:gaugeColor];
	[aCoder encodeObject:textColor];
	[aCoder encodeObject:titleFont];
}

@end


/****************************************************************************
  CHANGES:
   Version 0.3, March 19, 1995
  1. Fixed so it would draw with a transparent background correctly.
   Version 0.4, March 21, 1995
  2. Removed the lastRect and needRedraw ivars since as long as 
     drawSelf:inView: is called only when needed the face needs 
	 redrawing (from GaugeView) then drawing will be fairly efficient.
  3. Added drawHand:flipped: so you could easily customize the look
     of the gauge face and hand.
	 
 ****************************************************************************/
 