//-----------------------------------------------------------------------------------------------------------------
//		Written by Uwe Hoffmann Copyright 1995 by Uwe Hoffmann.
//				Version 1.0.  All rights reserved.
//
//		This notice may not be removed from this source code.
//-----------------------------------------------------------------------------------------------------------------

#import <Foundation/NSCoder.h>
#import "MiscSkipList.h"
#import "MiscSkipListEnumerator.h"

//-----------------------------------------------------------------------------------------------------------------
@implementation MiscSkipList

- init
{
	self = [super init];
	mutable = NO;
	head = allocNode([self zone], MAXLEVELS, nil);
	level = 1;
	nodeCount = 0;
	levelProvider = [[MiscRandomLevelProvider allocWithZone:[self zone]] init];
	return self;
}

- (void)dealloc
{
	[self removeAllObjects];
	freeNode(head);
	[levelProvider release];
	return [super dealloc];
}

- (void)encodeWithCoder:(NSCoder *)coder
{
	[super encodeWithCoder:coder];
	[coder encodeValueOfObjCType:"i" at:&mutable];
}

- initWithCoder:(NSCoder *)coder
{
    self = [super initWithCoder:coder];
	[coder decodeValueOfObjCType:"i" at:&mutable];
	head = allocNode([self zone], MAXLEVELS, nil);
	level = 1;
	nodeCount = 0;
	levelProvider = [[MiscRandomLevelProvider allocWithZone:[self zone]] init];
    return self;
}

- (void)insertObjectsFromSkipList:(MiscSkipList *)aList withZone:(NSZone *)aZone deep:(BOOL)aBool
{
	NSEnumerator *enumerator = [aList objectEnumerator];
	id object;
	
	if(aBool)
		while(object = [enumerator nextObject])
			[self insertObject:[object copyWithZone:aZone]];
	else
		while(object = [enumerator nextObject])
			[self insertObject:object];
	return;
}
	
- copyWithZone:(NSZone *)aZone
{
	MiscSkipList *theCopy;
	
	theCopy = [[MiscSkipList allocWithZone:aZone] init];
	[theCopy insertObjectsFromSkipList:self withZone:aZone deep:YES];
	return theCopy;
}

- mutableCopyWithZone:(NSZone *)aZone
{
	MiscSkipList *theCopy;
	
	theCopy = [[MiscSkipList allocWithZone:aZone] init];
	[theCopy insertObjectsFromSkipList:self withZone:aZone deep:NO];
	[theCopy setMutable:YES];
	return theCopy;
}
		 	
- (NSComparisonResult)compareNode:(SkipListNode *)p withObject:anObject
{
	return NSOrderedDescending;
}	

- (void)checkArgument:anObject
{
	return;
}

- (void)setMutable:(BOOL)aBool
{
	mutable = aBool;
}

- (unsigned)count
{
	return nodeCount;
}

- (NSEnumerator *)objectEnumerator
{
	MiscSkipListEnumerator *enumerator;
	
	enumerator = [[MiscSkipListEnumerator allocWithZone:[self zone]] initWithStart:head];
	return [enumerator autorelease];
}

- objectBefore:anObject
{
	SkipListNode *iterator;
	int i;
	
	if(!nodeCount)
		return nil;
	[self checkArgument:anObject];
	iterator = head;
	for(i = level - 1;i >= 0;i--)
		while([self compareNode:iterator->nodeForwards[i] withObject:anObject] == NSOrderedAscending)
			iterator = iterator->nodeForwards[i];
	return iterator->nodeObject;
}

- smallest
{
	if(!nodeCount)
		return nil;
	return head->nodeForwards[0]->nodeObject;
}

- (NSArray *)allObjectsOrderedSameAs:anObject
{
	SkipListNode *iterator;
	NSMutableArray *resultArray;
	int i;
	
	resultArray = [NSMutableArray array]; 
	if(!nodeCount)
		return resultArray;
	[self checkArgument:anObject];
	iterator = head;
	for(i = level - 1;i >= 0;i--)
		while([self compareNode:iterator->nodeForwards[i] withObject:anObject] == NSOrderedAscending)
			iterator = iterator->nodeForwards[i];
	iterator = iterator->nodeForwards[0];
	while([self compareNode:iterator withObject:anObject] == NSOrderedSame){
		[resultArray addObject:iterator->nodeObject];
		iterator = iterator->nodeForwards[0];
	}
	return resultArray;
}

- (void)removeAllObjectsOrderedSameAs:anObject;
{
	SkipListNode *update[MAXLEVELS];
	SkipListNode *iterator;
	int i;
	
	if(!nodeCount)
		return;
	[self checkArgument:anObject];
	iterator = head;
	for(i = level - 1;i >= 0;i--){
		while([self compareNode:iterator->nodeForwards[i] withObject:anObject] == NSOrderedAscending)
			iterator = iterator->nodeForwards[i];
		update[i] = iterator;
	}
	iterator = iterator->nodeForwards[0];
	while([self compareNode:iterator withObject:anObject] == NSOrderedSame){
		for(i = 0;i < level;i++)
			if(update[i]->nodeForwards[i] == iterator)
				update[i]->nodeForwards[i] = iterator->nodeForwards[i];
		[iterator->nodeObject release];
		freeNode(iterator);
		nodeCount--;
		iterator = update[0]->nodeForwards[0];
	}
	while(level > 1 && head->nodeForwards[level] == NULL)
		level--;
	return;
}
	
- (void)insertObject:anObject;
{
	SkipListNode *update[MAXLEVELS];
	SkipListNode *iterator, *newNode;
	int i, newLevel;
	
	[self checkArgument:anObject];
	[anObject retain];
	iterator = head;
	for(i = level - 1;i >= 0;i--){
		while([self compareNode:iterator->nodeForwards[i] withObject:anObject] == NSOrderedAscending)
			iterator = iterator->nodeForwards[i];
		update[i] = iterator;
	}
	iterator = iterator->nodeForwards[0];
	newLevel = [levelProvider randomLevel];
	if(newLevel > level){
		for(i = level;i < newLevel;i++)
			update[i] = head;
		level = newLevel;
	}
	newNode = allocNode([self zone], newLevel, anObject);
	nodeCount++;
	for(i = 0;i < newLevel;i++){
		newNode->nodeForwards[i] = update[i]->nodeForwards[i];
		update[i]->nodeForwards[i] = newNode;
	}
	return;
}

- (void)removeAllObjects
{
	SkipListNode *iterator, *deletedNode;			
	if(!nodeCount)
		return;
	iterator = head->nodeForwards[0];
	while(iterator){
		deletedNode = iterator;
		iterator = iterator->nodeForwards[0];
		 [deletedNode->nodeObject release];
		freeNode(deletedNode);
	}
	cutNode(head, MAXLEVELS);
	nodeCount = 0;
	level = 1;
	return;
}
	
@end
//-----------------------------------------------------------------------------------------------------------------


