/*
		Copyright (c) Uwe Hoffmann, 1995.
                  All Rights Reserved.

Filename: MiscUndoRecord.m
Author:   Uwe Hoffmann
Date:	  Jun 30, 1995

$Id: MiscUndoRecord.m,v 1.0 1995/06/30 11:22:00 Uwe Hoffmann Exp $
 $Log: MiscUndoRecord.m,v $
*/


#import <objc/objc-class.h>
#import "MiscUndoRecord.h"

#define BIT(x) (((unsigned int)1)<<(x))
#define BITS(from, to) ((((BIT(to) - 1)<<1) + 1) & !(BIT(from) - 1))

@implementation MiscUndoRecord
	/*"
	An instance of this class holds an NSInvocation instance used 
	to undo or redo other methods. In addition to this it provides 
	methods to specify  how and when the arguments of the NSInvocation 
	instance are freed.
	
	The methods #freeUndoArgs..., #freeUndoArg...At: specify whether the manager 
	should free its args when an 
	undo record is executed or when it exceeds the number of levels of undo. 
	The manager only frees non-object args. These methods set bits in the 
	freeArgsMask representing the argument position. Position 0-15 represent 
	the first 16 arguments to the method. Bits 0-15 in the freeArgsMask represent 
	args to be freed when the record is discarded, while Bits 16-31 represent args 
	to be freed when the record is executed. Methods use the macros BIT and BITS."*/ 

+ (MiscUndoRecord *)recordWithInvocation:(NSInvocation *)anInvocation
/*"Creates and returns an MiscUndoRecord instance holding anInvocation."*/
{
	MiscUndoRecord *theRecord;
	
	theRecord = [[MiscUndoRecord alloc] initWithInvocation:anInvocation];
	return [theRecord autorelease];
}

+ (MiscUndoRecord *)recordWithTarget:target selector:(SEL)selector,...
/*"Creates and returns an MiscUndoRecord instance holding an invocation with
   selector, target and arguments."*/
{
	NSMethodSignature *theSign;
	NSInvocation *theInv;
	// Method theMethod; // removed by DAY
	
	//theMethod = class_getInstanceMethod([target class], selector); // removed by DAY
	//theSign = [NSMethodSignature signatureWithObjCTypes:theMethod->method_types]; // removed by DAY
        theSign = [[target class] methodSignatureForSelector: selector]; // added by DAY
	theInv = [NSInvocation invocationWithMethodSignature:theSign];
	[theInv setTarget:target];
	[theInv setSelector: selector];
	return [self recordWithInvocation:theInv];
}

- initWithInvocation:(NSInvocation *)aInvocation
/*"Initializes a newly allocated record with anInvocation."*/
{
	[super init];
	invocation = [aInvocation retain];
	freeArgsMask = 0;
	return self;
}

- (void)dealloc
{
	NSMethodSignature *methSig;
	//unsigned i;
	//NSArgumentInfo argInfo;
	
	methSig = [invocation methodSignature];
	//for(i = 2; i < [methSig numberOfArguments]; i++){
	//	argInfo = [methSig argumentInfoAtIndex:i];
	//	...
	//}
	[invocation release];
	return [super dealloc];
}

- (void)dispatch
/*"Invokes the invocation and afterwards frees the arguments specified to be freed on dispatch."*/
{
	NSMethodSignature *methSig;
	//unsigned i;
	//NSArgumentInfo argInfo;
	
	[invocation invoke];
	methSig = [invocation methodSignature];
	//for(i = 2; i < [methSig numberOfArguments]; i++){
	//	argInfo = [methSig argumentInfoAtIndex:i];
	//	...
	//}
	return;
}
	
- (void)freeUndoArgsOnDiscard
/*"This method specifies whether the record should free all arguments from invocation when the record gets released."*/ 
{
	freeArgsMask |= BITS(0,15); 
	return;
}

- (void)freeUndoArgOnDiscardAt:(int)pos
/*"This method specifies whether the record should free the arguments from  invocation when the record gets released. The method sets bits 0 - 15 in the freeArgsMask representing the argument pos. Position 0-15 represent the first 16 arguments in invocation."*/
{
	if(pos < 0 || pos > 15)
		return;
	freeArgsMask |= BIT(pos);
	return;
}

- (void)freeUndoArgsOnExecute
/*"This method specifies whether the record should free all arguments from invocation after the invocation gets dispatched."*/
{
	freeArgsMask |= BITS(16,31); 
	return;
}

- (void)freeUndoArgOnExecuteAt:(int)pos
/*"This method specifies whether the record should free the arguments from  invocation after the invocation gets dispatched. The method sets bits 0 - 15 in the freeArgsMask representing the argument pos. Position 0-15 represent the first 16 arguments in invocation."*/
{
	if(pos < 0 || pos > 15)
		return;
	freeArgsMask |= BIT(pos + 16);
	return;
}

@end
