/*
**	SwitchView.m 
**	Copyright (C) 1995  David Slotnick 
**	 
**	This program is free software; you can redistribute it and/or 
**	modify it under the terms of the GNU General Public License as 
**	published by the Free Software Foundation; either version 1, or 
**	(at your option) any later version. 
**	 
**	This program is distributed in the hope that it will be useful, 
**	but WITHOUT ANY WARRANTY; without even the implied warranty of 
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
**	GNU General Public License for more details. 
**	 
**	You should have received a copy of the GNU General Public License 
**	along with this program; if not, write to the Free Software 
**	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

/***************************************************************************/
// RCS identification information
//
static char rcsID[] = "$Id: SwitchView.m,v 1.5 1995/03/31 21:26:06 davids Exp $";
static void __AvoidCompilerWarning(void) {if(!rcsID)__AvoidCompilerWarning();}

/***************************************************************************/
// Included Files
//
#import "SwitchView.h"

/***************************************************************************/
// Public Methods
//
@implementation SwitchView

/*
**	Standard initializer. 
*/
- initFrame:(const NXRect *)frameRect
{
	[super initFrame:frameRect];
	
	[self setTitle:"SwitchView"];
	views = [[NSMutableArray alloc] init];

	// add the first subView
	[self addView];
	[self setContentView:[self viewAtIndex:[self firstIndex]]];

	return self;
}

/*
**	Standard deallocator. 
*/
- free
{
	[views removeAllObjects];
	[views release];
	return [super free];
}

/*
**	Delegate methods. 
*/
- delegate { return delegate; }
- setDelegate:anObj { delegate = anObj; return self; }

/*
**	Adds a View to the end of the views Array. 
*/
- (void) addView
{
	// get permission from delegate
	if (!delegate || 
			([delegate respondsTo:@selector (switchViewWillAddView:)] &&
			[delegate switchViewWillAddView:self])) {

		View *newView;
	
		// insert and display a new View
		newView = [[View allocFromZone:[self zone]] initFrame:NULL];
		[views addObject:newView];
		[self displayViewAtIndex:[self lastIndex]];
	
		// notify delegate
		if ([delegate respondsTo:@selector (switchViewDidAddView:)]) {
			[delegate switchViewDidAddView:self];
		}
	}
}

/*
**	Inserts the View at the specified index, or raises an NSRangeException 
**	if the index is invalid. 
*/
- (void) insertViewAtIndex:(int)index
{
	// add new View and display
	if (![self isIndexValid:index]) {
	
		[NSException raise:NSRangeException 
			format:@"-[SwitchView insertViewAtIndex:]:: %d out of range",
				index];
				
	} else {
	
		// get permission from delegate
		if (!delegate || 
				([delegate respondsTo:
					@selector (switchView:willInsertViewAtIndex:)] &&
				[delegate switchView:self willInsertViewAtIndex:index])) {

			View *newView;
	
			/*
			**	Insert an empty View and redisplay the 
			**	_current_ View. The current View is 
			**	redisplayed because related methods 
			**	(-addView and  -deleteViewAtIndex:) 
			**	display a View in their implementation, 
			**	and clients of this class would probably 
			**	expect this method to do the same. (For 
			**	example, clients might be expecting to 
			**	catch the delegate messages from 
			**	-displayViewAtIndex:. 
			*/
			newView = [[View allocFromZone:[self zone]] initFrame:NULL];
			[views insertObject:newView atIndex:index];
			[self displayViewAtIndex:index];
				
			// notify delegate
			if ([delegate respondsTo:
					@selector (switchView:didInsertViewAtIndex:)]) {
				[delegate switchView:self didInsertViewAtIndex:index];
			}
		}
	}
}

/*
**	Displays the View at the specified index, or raises an NSRangeException 
**	if the index is invalid. 
*/
- (void) displayViewAtIndex:(int)index
{
	if (![self isIndexValid:index]) {
		
		[NSException raise:NSRangeException 
			format:@"-[SwitchView displayViewAtIndex:]:: %d out of range",
				index];

	} else {

		/*
		**	If the currently focused View is one of 
		**	my subviews (our one of my subviews' 
		**	subviews, etc.), the -setContentView: 
		**	message below will result in a "removed 
		**	a View from the View hierarchy that had 
		**	been lockFocus'ed" error.  So, if the 
		**	currently focused View is one of my 
		**	subviews, I resend this message to myself 
		**	every .1 seconds until the View relinquishes 
		**	its focus, and *then* I proceed with the 
		**	View switch. 
		*/
		if ([[NXApp focusView] isDescendantOf:self]) {
			[self perform:@selector (displayViewAtIndex:) 
				with:(void *)index
				afterDelay:100
				cancelPrevious:YES];
		
		/*
		**	The currently focused View isn't one of my 
		**	subviews, so get permission from the delegate. 
		*/
		} else {
			View 	*newContentView = [views objectAtIndex:index];
			View	*oldContentView = [self contentView];
			
			// get permission from delegate
			if ((oldContentView != newContentView) &&
				  (!delegate || 
					([delegate respondsTo:
						@selector (switchView:willDisplayViewAtIndex:)] &&
					[delegate switchView:self willDisplayViewAtIndex:index]))){
	
				// change content Views
				[self setContentView:nil];
				[self setContentView:newContentView];
				[oldContentView removeFromSuperview];
				
				// redisplay myself	
				[self display];
		
				// notify delegate
				if ([delegate respondsTo: 
						@selector (switchView:didDisplayViewAtIndex:)]) {
					[delegate switchView:self didDisplayViewAtIndex:index];
				}
			}
		}
	}
}

/*
**	Deletes the View at the specified index, or raises an NSRangeException
**	if the index is invalid or if there is only oneView in theSwitchView. 
*/
- (void) deleteViewAtIndex:(int)index
{
	// don't delete the last View
	if ([views count] == 1) {
	
		/*
		**	For some reason, the "@" operator doesn't 
		**	like strings concatenated by proximity 
		**  (e.g, "foo" "bar" --> "foobar"). 
		*/
		[NSException raise:NSRangeException 
			format:@"-[SwitchView deleteViewAtIndex:]:: View count == 1, cannot delete last View"];

	} else if (![self isIndexValid:index]) {

		[NSException raise:NSRangeException 
			format:@"-[SwitchView deleteViewAtIndex:]:: %d out of range",
				index];

	} else {
	
		// get permission from delegate
		if (!delegate || 
				([delegate respondsTo:
					@selector (switchView:willDeleteViewAtIndex:)] &&
				[delegate switchView:self willDeleteViewAtIndex:index])) {
				
			View 	*viewToRemove = [views objectAtIndex:index];
			BOOL	isLastView = (index == [self lastIndex]); 
			
			/*
			**	The Box implementation of -setContentView: 
			**	implementation messages the current 
			**	contentView in its implementation, so if 
			**	the View we're about to delete is the 
			**	current content View it should be detached 
			**	first. 
			*/
			if (viewToRemove == [self contentView]) {
				[self setContentView:nil];
			}

			// remove the view
			[viewToRemove removeFromSuperview];
			[views removeObjectAtIndex:index]; // frees the View
			
			/*
			**	Display the View moved to close the gap, 
			**	or the last View if the current last View 
			**	was just deleted. 
			*/
			if (isLastView) {
				[self displayViewAtIndex:[self lastIndex]];
			} else {
				[self displayViewAtIndex:index];
			}
						
			// redisplay myself
			[self display];

			// notify delegate
			if ([delegate respondsTo: 
					@selector (switchView:didDeleteViewAtIndex:)]) {
				[delegate switchView:self didDeleteViewAtIndex:index];
			}
		}
	}
}

/*
**	Return the index of the first and last Views. 
*/
- (int) firstIndex { return 0; }
- (int) lastIndex { return [views count] - 1; }

/*
**	Returns the View at the specified index, or raises an NSRangeException 
**	and returns nil if the index is invalid. 
*/
- (View *) viewAtIndex:(int)index 
{ 
	View *theView = nil;
	
	if (![self isIndexValid:index]) {
	
		[NSException raise:NSRangeException 
			format:@"-[SwitchView viewAtIndex:]:: %d out of range", index];
		
	} else {
		theView = [views objectAtIndex:index];
	}
	
	return theView;
}


/*
**	Returns the argument's preceding index, the index of the last 
**	View ([self lastIndex]) if the argument is the index of the first 
**	View ([self firstIndex]), or the index of the first View ([self 
**	firstIndex]) if the argument is an invalid index. 
*/
- (int) indexBefore:(int)index 
{
	int indexBefore;
	
	if ([self isIndexValid:index]) {
		if (index == [self firstIndex]) {
			indexBefore = [self lastIndex];
		} else {
			indexBefore = index - 1;
		}
		
	} else {
		indexBefore = [self firstIndex];
	}
	
	return indexBefore;
}

/*
**	Returns the argument's succeeding index, the index of the first 
**	View ([self firstIndex]) if the argument is the index of the last 
**	View ([self lastIndex]), or the index of the last View ([self 
**	lastIndex]) if the argument is an invalid index. 
*/
- (int) indexAfter:(int)index 
{
	int indexAfter;
	
	if ([self isIndexValid:index]) {
		if (index == [self lastIndex]) {
			indexAfter = [self firstIndex];
		} else {
			indexAfter = index + 1;
		}
		
	} else {
		indexAfter = [self lastIndex];
	}
	
	return indexAfter;
}

/*
**	Returns the number of Views in the SwitchView. 
*/
- (int) viewCount { return [views count]; }

/*
**	Returns the index of the currently visible View. 
*/
- (int) indexOfVisibleView
{ 
	int returnIndex;
	if (![self contentView]) {
		returnIndex = SV_NO_VISIBLE_VIEW;
	} else {		
		returnIndex = [views indexOfObject:[self contentView]];
	}
	return returnIndex;
}

/*
**	Checks index validity. 
*/
- (BOOL) isIndexValid:(int)index 
	{ return index >= [self firstIndex] && index <= [self lastIndex]; }

/*
**	Actions...
*/

- takeIntValueFrom:sender 
	{ [self displayViewAtIndex:[sender intValue]]; return self;  }
- takeStringValueFrom:sender 
	{ [self displayViewAtIndex:atoi ([(Cell *)sender stringValue])];
		 return self; }
- takeFloatValueFrom:sender 
	{ [self displayViewAtIndex:(int)[sender floatValue]]; return self; }
- takeDoubleValueFrom:sender 
	{ [self displayViewAtIndex:(int)[sender doubleValue]]; return self; }
- takeTagValueFrom:sender 
	{ [self displayViewAtIndex:(int)[sender tag]]; return self; }
- takeSelectedTagValueFrom:sender 
	{ [self displayViewAtIndex:(int)[sender selectedTag]]; return self; }


/*
**	Archiving Methods... 
*/

- awake 
{ 
	[self displayViewAtIndex:[views indexOfObject:[self contentView]]];
	return self;
}

- read:(NXTypedStream *)stream
{
	View 	*aView;
	int 	numViews;

	[super read:stream];

	NXReadTypes (stream, "i", &numViews);
	views = [[NSMutableArray alloc]	initWithCapacity:numViews];
	while (numViews--) {
		NXReadTypes (stream, "@", &aView);
		[views addObject:aView];
	}
	
	return self;
}

- write:(NXTypedStream *)stream
{
	NSEnumerator 	*enumerator = [views objectEnumerator];
	int				numViews = [views count];
	View 			*aView;

	[super write:stream];
	
	NXWriteTypes (stream, "i", &numViews);
	while (aView = [enumerator nextObject]) {
		NXWriteTypes (stream, "@", &aView);
	}

	return self;
}

@end
